import numpy.testing as npt
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.Cosmology.cosmology import Cosmology
import pytest
import numpy as np

class TestConeGeometry(object):

    def setup_method(self):

        self.arcsec = 2 * np.pi / 360 / 3600
        self.zlens = 1
        self.zsource = 2
        self.angle_diameter = 2/self.arcsec
        self.angle_radius = 0.5*self.angle_diameter

        H0 = 70
        omega_baryon = 0.0
        omega_DM = 0.3
        sigma8 = 0.82
        curvature = 'flat'
        ns = 0.9608
        cosmo_params = {'H0': H0, 'Om0': omega_baryon + omega_DM, 'Ob0': omega_baryon,
                      'sigma8': sigma8, 'ns': ns, 'curvature': curvature}
        self.cosmo = Cosmology(cosmo_kwargs=cosmo_params)

        self._angle_pad = 0.75
        self.geometry_double_cone = Geometry(self.cosmo, self.zlens, self.zsource, self.angle_diameter,
                                             'DOUBLE_CONE', self._angle_pad)

    def test_angle_scale(self):

        reduced_to_phys = self.geometry_double_cone.cosmo.D_A(0, self.zsource) / \
                          self.geometry_double_cone.cosmo.D_A(self.zlens, self.zsource)

        ratio_double_cone = reduced_to_phys * \
                            self.cosmo.D_A(self.zlens, self.zsource)/self.cosmo.D_A_z(self.zsource)
        angle_scale_zsource = 1 - self._angle_pad*ratio_double_cone

        npt.assert_almost_equal(self.geometry_double_cone.rendering_scale(self.zlens), 1.)
        npt.assert_almost_equal(self.geometry_double_cone.rendering_scale(self.zsource), angle_scale_zsource)

        npt.assert_almost_equal(self.geometry_double_cone.rendering_scale(self.zlens), 1)

    def test_distances_lensing(self):

        z = 0.3
        radius_physical = self.geometry_double_cone.angle_to_physicalradius(self.angle_radius, z)
        radius = self.geometry_double_cone.cosmo.D_A(0., z) * self.angle_radius * self.arcsec
        npt.assert_almost_equal(radius_physical, radius, 0)

        comoving_radius = self.geometry_double_cone.angle_to_comovingradius(self.angle_radius, z)
        npt.assert_almost_equal(comoving_radius, radius_physical * (1 + z), 3)

        z = 1
        radius_physical = self.geometry_double_cone.angle_to_physicalradius(self.angle_radius, z)
        radius = self.geometry_double_cone.cosmo.D_A(0., z) * self.angle_radius * self.arcsec
        npt.assert_almost_equal(radius_physical, radius, 0)

        comoving_radius = self.geometry_double_cone.angle_to_comovingradius(self.angle_radius, z)
        npt.assert_almost_equal(comoving_radius, radius_physical * (1+z), 3)

        z = 1.25
        radius_physical = self.geometry_double_cone.angle_to_physicalradius(self.angle_radius, z)
        D_dz = self.geometry_double_cone.cosmo.D_A(self.zlens, z)
        D_s = self.geometry_double_cone.cosmo.D_A(0, self.zsource)
        D_z = self.geometry_double_cone.cosmo.D_A(0, z)
        D_ds = self.geometry_double_cone.cosmo.D_A(self.zlens, self.zsource)

        rescale = 1 - self._angle_pad * D_dz * D_s / (D_z * D_ds)
        radius = self.geometry_double_cone.cosmo.D_A(0., z) * self.angle_radius * self.arcsec * rescale

        npt.assert_almost_equal(radius_physical, radius, 0)

        comoving_radius = self.geometry_double_cone.angle_to_comovingradius(self.angle_radius, z)
        npt.assert_almost_equal(comoving_radius, radius_physical * (1 + z), 3)


    def test_volume(self):

        cone_arcsec = 3
        radius = cone_arcsec*0.5
        angle_pad = 0.7
        zlens = 1.
        zsrc = 1.8
        geo = Geometry(self.cosmo, zlens, zsrc, cone_arcsec, 'DOUBLE_CONE', angle_pad=angle_pad)
        astropy = geo.cosmo.astropy

        delta_z = 1e-3

        dV_pyhalo = geo.volume_element_comoving(0.6, delta_z)
        dV = astropy.differential_comoving_volume(0.6).value

        dV_astropy = dV * delta_z
        steradian = np.pi * (radius * self.arcsec) ** 2
        npt.assert_almost_equal(dV_astropy * steradian, dV_pyhalo, 5)

        angle_scale = geo.rendering_scale(1.3)
        dV_pyhalo = geo.volume_element_comoving(1.3, delta_z)
        dV = astropy.differential_comoving_volume(1.3).value
        dV_astropy = dV * delta_z
        steradian = np.pi * (radius * angle_scale * self.arcsec) ** 2
        npt.assert_almost_equal(dV_astropy * steradian, dV_pyhalo, 5)

    def test_total_volume(self):

        cone_arcsec = 4
        radius_radians = cone_arcsec * 0.5 * self.cosmo.arcsec
        geo = Geometry(self.cosmo, 0.5, 1.5, cone_arcsec, 'DOUBLE_CONE', angle_pad=1.)
        volume_pyhalo = 0
        z = np.linspace(0.0, 1.5, 200)
        for i in range(0, len(z)-1):
            delta_z = z[i+1] - z[i]
            volume_pyhalo += geo.volume_element_comoving(z[i], delta_z)

        ds = self.cosmo.D_C_z(1.5)
        dz = self.cosmo.D_C_z(0.5)
        volume_true = 1./3 * np.pi * radius_radians ** 2 * dz ** 2 * ds
        npt.assert_almost_equal(volume_true, volume_pyhalo, 3)

    def test_cone(self):

        cone_arcsec = 4
        radius_radians = cone_arcsec * 0.5 * self.cosmo.arcsec
        geo = Geometry(self.cosmo, 0.5, 1.5, cone_arcsec, 'CONE', angle_pad=1.)
        volume_pyhalo = 0
        z = np.linspace(0.0, 1.5, 200)
        for i in range(0, len(z)-1):
            delta_z = z[i+1] - z[i]
            volume_pyhalo += geo.volume_element_comoving(z[i], delta_z)

        ds = self.cosmo.D_C_z(1.5)
        volume_true = 1./3 * np.pi * radius_radians ** 2 * ds ** 3
        npt.assert_almost_equal(volume_true, volume_pyhalo, 3)


if __name__ == '__main__':
      pytest.main()



import numpy as np
from scipy.integrate import quad


def integrate_power_law_analytic(norm, m_low, m_high, n, plaw_index):
    """
    Numerically integrates a power law profile
    :param norm:
    :param m_low:
    :param m_high:
    :param n:
    :param plaw_index:
    :return:
    """
    factor = n + 1 + plaw_index
    if factor == 0:
        integral = np.log(m_high / m_low)
    else:
        integral = (m_high ** factor - m_low ** factor) / factor
    return norm * integral

def integrate_power_law_quad(norm, m_low, m_high, n, plaw_index, turnover_function, kwargs_turnover):

    """
    Numerically integrates a power law profile with an arbitrary turnover or shape function
    """

    def _integrand_wdm(m):
        return norm * m ** (n + plaw_index) * turnover_function(m, **kwargs_turnover)
    moment = quad(_integrand_wdm, m_low, m_high)[0]
    return moment



import pytest
from pyHalo.single_realization import SingleHalo
from pyHalo.realization_extensions import RealizationExtensions, corr_kappa_with_mask, xi_l, xi_l_to_Pk_l, fit_correlation_multipole
from pyHalo.Cosmology.cosmology import Cosmology
from scipy.interpolate import interp1d
import numpy.testing as npt
import numpy as np
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from pyHalo.Halos.tidal_truncation import TruncationRoche, TruncationRN
from pyHalo.Halos.lens_cosmo import LensCosmo
from scipy.special import eval_chebyt


class TestRealizationExtensions(object):

    def test_toSIDM(self):

        halo_mass = 10 ** 8
        x = 0.5
        y = 1.0
        z = 0.5
        zlens = 0.5
        zsource = 2.0
        astropy_instance = Cosmology().astropy
        lens_cosmo = LensCosmo(zlens, zsource)
        truncation_class = TruncationRN(lens_cosmo, 100.0)
        concentration_class = ConcentrationDiemerJoyce(astropy_instance, scatter=False)

        class DummyCrossSection(object):

            def effective_cross_section(self, v):
                return 100.0

        kwargs_halo_model = {'concentration_model': concentration_class,
                             'truncation_model': truncation_class,
                             'kwargs_density_profile': {'evaluate_mc_at_zlens': True}}
        cross_section = DummyCrossSection()
        mdef = 'TNFW'
        single_halo = SingleHalo(halo_mass, x, y, mdef, z, zlens, zsource, subhalo_flag=False,
                                 kwargs_halo_model=kwargs_halo_model)
        ext = RealizationExtensions(single_halo)
        single_halo_sidm_1 = ext.toSIDM(cross_section)

        single_subhalo = SingleHalo(halo_mass, x, y, mdef, z, zlens, zsource, subhalo_flag=True,
                                 kwargs_halo_model=kwargs_halo_model)
        ext = RealizationExtensions(single_subhalo)
        single_subhalo_sidm_1 = ext.toSIDM(cross_section)

        mdef = 'TNFWC_SIDM'
        kwargs_halo_model['kwargs_density_profile']['SIDM_CROSS_SECTION'] = cross_section
        single_halo_sidm_2 = SingleHalo(halo_mass, x, y, mdef, z, zlens, zsource, subhalo_flag=False,
                                 kwargs_halo_model=kwargs_halo_model)
        single_subhalo_sidm_2 = SingleHalo(halo_mass, x, y, mdef, z, zlens, zsource, subhalo_flag=True,
                                 kwargs_halo_model=kwargs_halo_model)
        kwargs_lens_1 = single_halo_sidm_1.lensing_quantities(add_mass_sheet_correction=False)[2][0]
        kwargs_lens_2 = single_halo_sidm_2.lensing_quantities(add_mass_sheet_correction=False)[2][0]

        kwargs_lens_3 = single_subhalo_sidm_1.lensing_quantities(add_mass_sheet_correction=False)[2][0]
        kwargs_lens_4 = single_subhalo_sidm_2.lensing_quantities(add_mass_sheet_correction=False)[2][0]

        for kw in kwargs_lens_1.keys():
            npt.assert_almost_equal(kwargs_lens_1[kw], kwargs_lens_2[kw])
            npt.assert_almost_equal(kwargs_lens_4[kw], kwargs_lens_3[kw])

    def test_add_core_collapsed_halos(self):

        halo_mass = 10 ** 8
        x = 0.5
        y = 1.0
        mdef = 'TNFW'
        z = 0.5
        zlens = 0.5
        zsource = 2.0
        astropy_instance = Cosmology().astropy
        truncation_class = TruncationRoche()
        concentration_class = ConcentrationDiemerJoyce(astropy_instance)
        kwargs_halo_model = {'concentration_model_field_halos': concentration_class,
                                    'truncation_model_field_halos': truncation_class,
                                    'kwargs_density_profile': {}}
        single_halo = SingleHalo(halo_mass, x, y, mdef, z, zlens, zsource, subhalo_flag=False,
                 kwargs_halo_model=kwargs_halo_model)
        ext = RealizationExtensions(single_halo)
        new = ext.add_core_collapsed_halos([0], log_slope_halo=3.1, x_core_halo=0.05)
        lens_model_list, zlist, kwargs_halo, _ = new.lensing_quantities()
        npt.assert_string_equal(lens_model_list[0], 'SPL_CORE')
        npt.assert_equal(kwargs_halo[0]['gamma'], 3.1)

    def test_collapse_profile(self):

        halo_mass = 10 ** 9
        x = 0.5
        y = 1.0
        z = 0.5
        zlens = 0.5
        zsource = 2.0
        astropy_instance = Cosmology().astropy
        truncation_class = TruncationRoche()
        concentration_class = ConcentrationDiemerJoyce(astropy_instance, scatter=False)

        kwargs_density_profile_spl_core = {'log_slope_halo': 3.2, 'x_core_halo': 0.05, 'x_match': 'c'}
        kwargs_halo_model = {'concentration_model_field_halos': concentration_class,
                             'truncation_model_field_halos': truncation_class,
                             'kwargs_density_profile': kwargs_density_profile_spl_core}
        realization_1 = SingleHalo(halo_mass, x, y, 'SPL_CORE', z, zlens, zsource, subhalo_flag=False,
                 kwargs_halo_model=kwargs_halo_model)

        kwargs_halo_model_2 = {'concentration_model_field_halos': concentration_class,
                             'truncation_model_field_halos': truncation_class,
                             'kwargs_density_profile': {}}
        realization_2 = SingleHalo(halo_mass, x, y, 'TNFW', z, zlens, zsource, subhalo_flag=False,
                                   kwargs_halo_model=kwargs_halo_model_2)

        ext = RealizationExtensions(realization_2)
        realization_collapsed = ext.add_core_collapsed_halos([0], halo_profile='SPL_CORE', **kwargs_density_profile_spl_core)
        lens_model_list, zlist, kwargs_halo, _ = realization_collapsed.lensing_quantities()
        lens_model_list_1, zlist_1, kwargs_halo_1, _ = realization_1.lensing_quantities()
        npt.assert_string_equal(lens_model_list[0], lens_model_list_1[0])
        npt.assert_equal(zlist[0], zlist_1[0])
        npt.assert_equal(kwargs_halo, kwargs_halo_1)

        kwargs_density_profile_gnfw = {'gamma_inner': 1.9, 'gamma_outer': 2.85, 'x_match': 'c'}
        kwargs_halo_model = {'concentration_model_field_halos': concentration_class,
                             'truncation_model_field_halos': truncation_class,
                             'kwargs_density_profile': kwargs_density_profile_gnfw}
        realization_1 = SingleHalo(halo_mass, x, y, 'GNFW', z, zlens, zsource, subhalo_flag=False,
                                   kwargs_halo_model=kwargs_halo_model)

        kwargs_halo_model_2 = {'concentration_model_field_halos': concentration_class,
                               'truncation_model_field_halos': truncation_class,
                               'kwargs_density_profile': {}}
        realization_2 = SingleHalo(halo_mass, x, y, 'TNFW', z, zlens, zsource, subhalo_flag=False,
                                   kwargs_halo_model=kwargs_halo_model_2)

        ext = RealizationExtensions(realization_2)
        realization_collapsed = ext.add_core_collapsed_halos([0], halo_profile='GNFW', **kwargs_density_profile_gnfw)
        lens_model_list, zlist, kwargs_halo, _ = realization_collapsed.lensing_quantities()
        lens_model_list_1, zlist_1, kwargs_halo_1, _ = realization_1.lensing_quantities()
        npt.assert_string_equal(lens_model_list[0], lens_model_list_1[0])
        npt.assert_equal(zlist[0], zlist_1[0])
        npt.assert_equal(kwargs_halo, kwargs_halo_1)

    def test_collapse_by_mass(self):

        cosmo = Cosmology()
        m_list = 10**np.random.uniform(6, 10, 1000)
        astropy_instance = Cosmology().astropy
        truncation_class = TruncationRoche()
        concentration_class = ConcentrationDiemerJoyce(astropy_instance, scatter=False)
        kwargs_halo_model = {'concentration_model': concentration_class,
                             'truncation_model': truncation_class,
                             'kwargs_density_profile': {}}
        lens_cosmo = LensCosmo(0.5, 2.0, cosmo)
        realization = SingleHalo(m_list[0], 0.0, 0.0, 'TNFW', 0.5, 0.5, 1.5,
                                 subhalo_flag=True, kwargs_halo_model=kwargs_halo_model, lens_cosmo=lens_cosmo)

        for mi in m_list[1:]:
            single_halo = SingleHalo(mi, 0.0, 0.0, 'TNFW', 0.5, 0.5, 1.5,
                                 subhalo_flag=True, kwargs_halo_model=kwargs_halo_model, lens_cosmo=lens_cosmo)
            realization = realization.join(single_halo)
            single_halo = SingleHalo(mi, 0.0, 0.0, 'TNFW', 0.5, 0.5, 1.5,
                                 subhalo_flag=False, kwargs_halo_model=kwargs_halo_model, lens_cosmo=lens_cosmo)
            realization = realization.join(single_halo)

        ext = RealizationExtensions(realization)

        mass_range_subs = [[6, 8], [8, 10]]
        mass_range_field = [[6, 8], [8, 10]]
        p_subs = [0.3, 0.9]
        p_field = [0.8, 0.25]
        kwargs_halo = {'log_slope_halo': 3, 'x_core_halo': 0.05}
        inds_collapsed = ext.core_collapse_by_mass(mass_range_subs, mass_range_field,
                              p_subs, p_field)
        realization_collapsed = ext.add_core_collapsed_halos(inds_collapsed, **kwargs_halo)


        i_subs_collapsed_1 = 0
        i_subs_1 = 0
        i_field_collapsed_1 = 0
        i_field_1 = 0
        i_subs_collapsed_2 = 0
        i_subs_2 = 0
        i_field_collapsed_2 = 0
        i_field_2 = 0
        for halo in realization_collapsed.halos:

            if halo.is_subhalo:
                if halo.mass < 10 ** 8:
                    i_subs_1 += 1
                    if halo.mdef == 'SPL_CORE':
                        i_subs_collapsed_1 += 1
                else:
                    i_subs_2 += 1
                    if halo.mdef == 'SPL_CORE':
                        i_subs_collapsed_2 += 1
            else:
                if halo.mass < 10 ** 8:
                    i_field_1 += 1
                    if halo.mdef == 'SPL_CORE':
                        i_field_collapsed_1 += 1
                else:
                    i_field_2 += 1
                    if halo.mdef == 'SPL_CORE':
                        i_field_collapsed_2 += 1

        npt.assert_almost_equal(abs(p_subs[0] - i_subs_collapsed_1 / i_subs_1), 0, 1)
        npt.assert_almost_equal(abs(p_subs[1] - i_subs_collapsed_2 / i_subs_2), 0, 1)
        npt.assert_almost_equal(abs(p_field[0] - i_field_collapsed_1 / i_field_1), 0, 1)
        npt.assert_almost_equal(abs(p_field[1] - i_field_collapsed_2 / i_field_2), 0, 1)

    def test_collapse_by_mass_redshift(self):

        cosmo = Cosmology()
        m_list = 10 ** np.random.uniform(6, 10, 1000)
        astropy_instance = Cosmology().astropy
        truncation_class = TruncationRoche()
        concentration_class = ConcentrationDiemerJoyce(astropy_instance, scatter=False)
        kwargs_halo_model = {'concentration_model': concentration_class,
                             'truncation_model': truncation_class,
                             'kwargs_density_profile': {}}
        lens_cosmo = LensCosmo(0.5, 2.0, cosmo)
        realization = SingleHalo(m_list[0], 0.0, 0.0, 'TNFW', 0.5, 0.5, 1.5,
                                 subhalo_flag=True, kwargs_halo_model=kwargs_halo_model, lens_cosmo=lens_cosmo)

        for mi in m_list[1:]:
            single_halo = SingleHalo(mi, 0.0, 0.0, 'TNFW', 0.5, 0.5, 1.5,
                                     subhalo_flag=True, kwargs_halo_model=kwargs_halo_model, lens_cosmo=lens_cosmo)
            realization = realization.join(single_halo)
            single_halo = SingleHalo(mi, 0.0, 0.0, 'TNFW', 0.5, 0.5, 1.5,
                                     subhalo_flag=False, kwargs_halo_model=kwargs_halo_model, lens_cosmo=lens_cosmo)
            realization = realization.join(single_halo)

        ext = RealizationExtensions(realization)

        mass_range_subs = [[6, 8], [8, 10]]
        mass_range_field = [[6, 8], [8, 10]]

        custom_func_1 = lambda z: 0.3
        custom_func_2 = lambda z: 0.9
        custom_func_3 = lambda z: 0.8
        custom_func_4 = lambda z: 0.25
        z_eval = 0.5
        p_subs = [custom_func_1, custom_func_2]
        p_field = [custom_func_3, custom_func_4]
        kwargs_sub = [{}, {}]
        kwargs_field = [{}, {}]
        kwargs_halo = {'log_slope_halo': -3, 'x_core_halo': 0.05}
        inds_collapsed = ext.core_collapse_by_mass(mass_range_subs, mass_range_field,
                              p_subs, p_field, kwargs_sub, kwargs_field)
        realization_collapsed = ext.add_core_collapsed_halos(inds_collapsed, **kwargs_halo)

        i_subs_collapsed_1 = 0
        i_subs_1 = 0
        i_field_collapsed_1 = 0
        i_field_1 = 0
        i_subs_collapsed_2 = 0
        i_subs_2 = 0
        i_field_collapsed_2 = 0
        i_field_2 = 0
        for halo in realization_collapsed.halos:

            if halo.is_subhalo:
                if halo.mass < 10 ** 8:
                    i_subs_1 += 1
                    if halo.mdef == 'SPL_CORE':
                        i_subs_collapsed_1 += 1
                else:
                    i_subs_2 += 1
                    if halo.mdef == 'SPL_CORE':
                        i_subs_collapsed_2 += 1
            else:
                if halo.mass < 10 ** 8:
                    i_field_1 += 1
                    if halo.mdef == 'SPL_CORE':
                        i_field_collapsed_1 += 1
                else:
                    i_field_2 += 1
                    if halo.mdef == 'SPL_CORE':
                        i_field_collapsed_2 += 1

        npt.assert_almost_equal(abs(p_subs[0](z_eval) - i_subs_collapsed_1 / i_subs_1), 0, 1)
        npt.assert_almost_equal(abs(p_subs[1](z_eval) - i_subs_collapsed_2 / i_subs_2), 0, 1)
        npt.assert_almost_equal(abs(p_field[0](z_eval) - i_field_collapsed_1 / i_field_1), 0, 1)
        npt.assert_almost_equal(abs(p_field[1](z_eval) - i_field_collapsed_2 / i_field_2), 0, 1)

    def test_add_ULDM_fluctuations(self):

        cosmo = Cosmology()
        astropy_instance = cosmo.astropy
        truncation_class = TruncationRoche()
        concentration_class = ConcentrationDiemerJoyce(astropy_instance, scatter=False)
        kwargs_halo_model = {'concentration_model': concentration_class,
                             'truncation_model': truncation_class,
                             'kwargs_density_profile': {}}
        lens_cosmo = LensCosmo(0.5, 2.0, cosmo)
        single_halo = SingleHalo(10 ** 8, 0.0, 0.0, 'TNFW', 0.5, 0.5, 1.5, 100.0,
                                 subhalo_flag=True, kwargs_halo_model=kwargs_halo_model, lens_cosmo=lens_cosmo)
        ext = RealizationExtensions(single_halo)
        wavelength=0.6 #kpc, correpsonds to m=10^-22 eV for ULDM
        amp_var = 0.04 #in convergence units
        fluc_var = wavelength
        fluc_size = 0.1
        n_cut = 1e4

        # apeture
        x_images = np.array([-0.347, -0.734, -1.096, 0.207])
        y_images = np.array([ 0.964,  0.649, -0.079, -0.148])
        args_aperture = {'x_images':x_images,'y_images':y_images,'aperture':0.25}
        new = ext.add_ULDM_fluctuations(wavelength,amp_var,fluc_size,fluc_var,n_cut,shape='aperture',args=args_aperture)
        _ = new.lensing_quantities()

        #ring
        args_ring = {'rmin':0.95,'rmax':1.05}
        new = ext.add_ULDM_fluctuations(wavelength,amp_var,fluc_size,fluc_var,n_cut,shape='ring',args=args_ring)
        _ = new.lensing_quantities()

        #ellipse
        args_ellipse = {'amin':0.8,'amax':1.7,'bmin':0.4,'bmax':1.2,'angle':np.pi/4}
        new = ext.add_ULDM_fluctuations(wavelength,amp_var,fluc_size,fluc_var,n_cut,shape='ellipse',args=args_ellipse)
        _ = new.lensing_quantities()

    def test_add_pbh(self):

        cosmo = Cosmology()
        astropy_instance = cosmo.astropy
        lens_cosmo = LensCosmo(0.5, 1.5, cosmo)
        truncation_class = TruncationRN(lens_cosmo)
        concentration_class = ConcentrationDiemerJoyce(astropy_instance, scatter=False)
        kwargs_halo_model = {'concentration_model': concentration_class,
                             'truncation_model': truncation_class,
                             'kwargs_density_profile': {}}

        realization = SingleHalo(10**9, 0.9, 0.1, 'TNFW', 0.5, 0.5, 1.5,
                                 subhalo_flag=False, kwargs_halo_model=kwargs_halo_model, lens_cosmo=lens_cosmo)
        zlist = [0.2, 0.4, 0.6]
        rmax = 0.4
        main_halo_coord_x = []
        main_halo_coord_y = []
        for i, zi in enumerate(zlist):
            nhalos = np.random.randint(1, 5)
            _x, _y = [], []
            for j in range(0, nhalos):
                mi = np.random.uniform(8,  9)
                xi = np.random.uniform(-0.2, 0.2)
                yi = np.random.uniform(-0.2, 0.2)
                _x.append(xi)
                _y.append(yi)
                single_halo = SingleHalo(10 ** mi, xi, yi, 'TNFW', zi, 0.5, 1.5, subhalo_flag=False,
                                     kwargs_halo_model=kwargs_halo_model, lens_cosmo=lens_cosmo)
                realization = realization.join(single_halo)
            main_halo_coord_x.append(_x)
            main_halo_coord_y.append(_y)

        lens_model_list_init, _, kwargs_init, _ = realization.lensing_quantities()
        ext = RealizationExtensions(realization)
        mass_fraction = 0.1
        kwargs_mass_function = {'mass_function_type': 'DELTA', 'logM': 5., 'mass_fraction': 0.5}
        fraction_in_halos = 0.9

        _zlist = np.round(np.arange(0.00, 1.02, 0.02), 2)
        x_image = [0.] * len(_zlist)
        y_image = [0.] * len(_zlist)
        cosmo = Cosmology()
        dlist = [cosmo.D_C_transverse(zi) for zi in _zlist]
        x_image_interp_list = [interp1d(dlist, x_image)]
        y_image_interp_list = [interp1d(dlist, y_image)]

        r_array = np.zeros(len(x_image_interp_list))
        r_array[0:] = rmax

        pbh_realization = ext.add_primordial_black_holes(mass_fraction, kwargs_mass_function,
                                                         fraction_in_halos,
                                                         x_image_interp_list,
                                                         y_image_interp_list,
                                                         r_array)

        # import matplotlib.pyplot as plt
        # from pyHalo.single_realization import realization_at_z
        # for j, zi in enumerate(zlist):
        #     real = realization_at_z(pbh_realization, zi)[0]
        #     plt.scatter(real.x, real.y)
        #     plt.scatter(main_halo_coord_x[j], main_halo_coord_y[j], color='r')
        #     plt.show()
        #     a=input('continue')

        for i, halo in enumerate(pbh_realization.halos):

            condition1 = 'PT_MASS' == halo.mdef
            condition2 = 'TNFW' == halo.mdef
            npt.assert_equal(np.logical_or(condition1, condition2), True)

# class TestCorrelationComputation(object):
#
#     def setup_method(self):
#         pass
#
#     def test_corr_kappa_with_mask(self):
#         npix = 500
#         window_size = 4
#         delta_pix = window_size/npix
#         mu = np.linspace(-1, 1, 100)
#         r = np.logspace(np.log10(2*10**(-2)), -0.3, num=100, endpoint=True)
#         _R = np.linspace(-window_size/2, window_size/2, npix)
#         XX, YY = np.meshgrid(_R, _R)
#
#         def kappa_GRF(delta_pix, num_pix, alpha):
#             #Generating Gaussian random field kappa map
#             noise = (delta_pix**2)*np.fft.fft2(np.random.normal(size=(num_pix,num_pix)))
#             fftind = 2.0*np.pi*np.fft.fftfreq(num_pix, d=delta_pix)
#             kxi,kyi = np.meshgrid(fftind,fftind)
#             kvnorm2 = kxi**2 + kyi**2 + 1e-10
#             amplitude = np.sqrt((kvnorm2*delta_pix**2)**(-alpha/2))
#             kappa = np.fft.ifft2(amplitude*noise)/delta_pix**2
#
#             return kappa.real - np.mean(kappa.real)
#
#         alpha = 1
#         kappa = kappa_GRF(delta_pix, npix, alpha)
#
#         corr = corr_kappa_with_mask(kappa, window_size, r, mu, apply_mask = False, r_min = 0, r_max = None, normalization = False)
#         corr_mask = corr_kappa_with_mask(kappa, window_size, r, mu, apply_mask = True, r_min = 0.5, r_max = None, normalization = False)
#         corr_mask_ann = corr_kappa_with_mask(kappa, window_size, r, mu, apply_mask = True, r_min = 0.5, r_max = 1.5, normalization = False)
#         corr_norm = corr_kappa_with_mask(kappa, window_size, r, mu, apply_mask = False, r_min = 0, r_max = None, normalization = True)
#         corr_mask_norm = corr_kappa_with_mask(kappa, window_size, r, mu, apply_mask = True, r_min = 0.5, r_max = None, normalization = True)
#         corr_mask_ann_norm = corr_kappa_with_mask(kappa, window_size, r, mu, apply_mask = True, r_min = 0.5, r_max = 1.5, normalization = True)
#
#         xi_0_real = delta_pix**(2-alpha)/(2*np.pi*r)
#
#         mu_grid = np.tile(mu, (r.shape[0], 1))
#         T_l_grid = eval_chebyt(0, mu_grid)
#         xi_l_grid = np.transpose([xi_0_real] *mu.shape[0])
#
#         corr_real = xi_l_grid*T_l_grid
#         corr_real_norm = np.linalg.norm(corr_real, 1)*corr_real
#
#         npt.assert_array_almost_equal(corr_real, corr, decimal=2)
#         npt.assert_array_almost_equal(corr_real, corr_mask, decimal=2)
#         npt.assert_array_almost_equal(corr_real, corr_mask_ann, decimal=2)
#         npt.assert_array_almost_equal(corr_real_norm,corr_norm, 1)
#         npt.assert_array_almost_equal(corr_real_norm,corr_mask_norm, 1)
#         npt.assert_array_almost_equal(corr_real_norm,corr_mask_ann_norm, 1)
#
#     def test_xi_l(self):
#         mu = np.linspace(-1, 1, 100)
#         r = np.logspace(-3, -0.3, num=100, endpoint=True)
#         xi_0_real = np.ones(r.shape[0])
#         corr = np.ones((r.shape[0], mu.shape[0]))
#         r, xi_0 = xi_l(0, corr, r, mu)
#         npt.assert_almost_equal(xi_0_real, xi_0)
# 
#     def test_xi_l_to_Pk_l(self):
#         l = 0
#         x = np.logspace(-3, 3, num=60, endpoint=False)
#         F = 1 / (1 + x*x)**1.5
#         y, G_Hankel = xi_l_to_Pk_l(x, F, l = 0)
#         G = (2*np.pi*(-1j)**l) * np.exp(-y)  # this is the actual Hankel transform of the function F.
#         npt.assert_almost_equal(G, G_Hankel)
#
#     def test_fit_correlation_multipole(self):
#         r = np.logspace(-1, 2, num=100, endpoint=True)
#         As = 5
#         n = -3
#         r_min, r_max = 0, 50
#         r_pivot = (r_min + r_max)/2
#         func_real = As*(r/r_pivot)**n
#
#         As_fit, n_fit = fit_correlation_multipole(r, func_real, r_min, r_max)
#         npt.assert_array_almost_equal(As, As_fit)
#         npt.assert_array_almost_equal(n, n_fit)


if __name__ == '__main__':
      pytest.main()



import numpy as np
import h5py
import re

"""
AUTHOR: Xiaolong Du

Read the halo information from Galacticus outputs.

Default units:

Mass            Msun
Time            Gyr
Length          Mpc
Velocity        km/s

Available halo properties:

time            time since the big bang

posX            x position of a halo with respect to the host halo
posY            y position of a halo with respect to the host halo
posZ            z position of a halo with respect to the host halo
velX            x velocity of a halo with respect to the host halo
velY            y velocity of a halo with respect to the host halo
velZ            z velocity of a halo with respect to the host halo
Rorbit          Distance to the host center

Minfall         Mass at infall
Mbound          Bound mass
Rbound          Radius eonclosing the bound mass
Mvir            Virial mass defined with a specified virial density contrast definition
                (for subhalo this may be slightly different Mbound)
Vmax            Maximum circualr velocity
Rmax            Radius where the circular velocity reaches its maximum
Rtidal          Tidal radius
Rapo            Apocenter distance
Rperi           Pericenter distance

MHost           Virial mass of the host halo at each output
RHost           Virial radius of the host halo at each output
RsHost          Scale radius of the host halo at each output
RvirSat0        Virial radius of the subahlo at infall
RsSat0          Scale radius of the subahlo at infall
cSat0           Concentration of the subhalo at infall

## Density profile
densityHost(i)  Return the density profile of host halo at i-th snapshot. Note that i starts with 0
density    (i)  Return the density profile of subhalo at i-th snapshot. Note that i starts with 0
densityProj(i)  Return the projected density of subhalo at i-th snapshot. Note that i starts with 0

"""


class GalacticusDataSingle_Du(object):

    def __init__(self, fileName):
        """
        This class reads and stores information from a single galacitucs run
        :param fileName: the file path for the hdf5file output by galacticus
        """
        # Open the data file.
        self.fileRead = h5py.File(fileName, 'r')
        self.NOutput = len(self.fileRead["Outputs"].keys())
        self.fileName = fileName

        self.time = np.zeros(self.NOutput)
        self.posX = np.zeros(self.NOutput)
        self.posY = np.zeros(self.NOutput)
        self.posZ = np.zeros(self.NOutput)
        self.velX = np.zeros(self.NOutput)
        self.velY = np.zeros(self.NOutput)
        self.velZ = np.zeros(self.NOutput)
        self.Rorbit = np.zeros(self.NOutput)
        self.Minfall = np.zeros(self.NOutput)
        self.Mbound = np.zeros(self.NOutput)
        self.Rbound = np.zeros(self.NOutput)
        self.Mvir = np.zeros(self.NOutput)
        self.Vmax = np.zeros(self.NOutput)
        self.Rmax = np.zeros(self.NOutput)
        self.Rtidal = np.zeros(self.NOutput)
        self.Rapo = np.zeros(self.NOutput)
        self.Rperi = np.zeros(self.NOutput)
        self.MHost = np.zeros(self.NOutput)
        self.RHost = np.zeros(self.NOutput)
        self.RsHost = np.zeros(self.NOutput)

        for i in range(self.NOutput):
            basicMass = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/basicMass"][:]
            indexSat = np.argmin(basicMass)
            indexHost = np.argmax(basicMass)

            self.time[i] = self.fileRead["Outputs/Output" + str(i + 1)].attrs["outputTime"]
            self.posX[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/positionOrbitalX"][indexSat]
            self.posY[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/positionOrbitalY"][indexSat]
            self.posZ[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/positionOrbitalZ"][indexSat]
            self.velX[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/velocityOrbitalX"][indexSat]
            self.velY[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/velocityOrbitalY"][indexSat]
            self.velZ[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/velocityOrbitalZ"][indexSat]
            self.Minfall[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/massBasic"][indexSat]
            self.Mbound[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/massBound"][indexSat]
            self.Rbound[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/satelliteRadiusBoundMass"][
                indexSat]
            self.Mvir[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/massHaloEnclosedCurrent"][indexSat]
            self.Vmax[i] = \
            self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/darkMatterProfileDMOVelocityMaximum"][indexSat]
            self.Rmax[i] = \
            self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/darkMatterProfileDMORadiusVelocityMaximum"][
                indexSat]
            self.Rtidal[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/satelliteRadiusTidal"][indexSat]

            self.Rapo[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/satelliteApocenterRadius"][indexSat]
            self.Rperi[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/satellitePericenterRadius"][
                indexSat]
            self.MHost[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/massBasic"][indexHost]
            self.RHost[i] = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/darkMatterOnlyRadiusVirial"][
                indexHost]
            self.RsHost[i] = self.fileRead["Outputs/Output" + str(1) + "/nodeData/darkMatterProfileScale"][indexHost]

        self.Rorbit = np.sqrt(self.posX ** 2 + self.posY ** 2 + self.posZ ** 2)
        self.Vorbit = np.sqrt(self.velX ** 2 + self.velY ** 2 + self.velZ ** 2)
        self.Lz = self.posX * self.velY - self.posY * self.velX

        basicMass = self.fileRead["Outputs/Output" + str(1) + "/nodeData/basicMass"][:]
        indexSat = np.argmin(basicMass)
        self.RvirSat0 = self.fileRead["Outputs/Output" + str(1) + "/nodeData/darkMatterOnlyRadiusVirial"][indexSat]
        self.RsSat0 = self.fileRead["Outputs/Output" + str(1) + "/nodeData/darkMatterProfileScale"][indexSat]
        self.cSat0 = self.RvirSat0 / self.RsSat0

        radii = np.array([])
        radiiName = []
        radiiColumns = self.fileRead["Outputs/Output" + str(1) + "/nodeData/densityProfileColumns"][:]
        for dataName in radiiColumns:
            split = re.split(r":", dataName.decode())
            radii = np.append(radii, float(split[-1]))
            radiiName.append(split[-1])

        self.radiiRatio = radii
        self.radiiSat = radii * self.RvirSat0
        self.radiiName = radiiName
        self.NRadii = len(radii)

        self.fileRead.close()

    def __del__(self):
        self.fileRead.close()

    def densityHost(self, i):
        # Return the density profile of the host.
        self.fileRead = h5py.File(self.fileName, 'r')
        basicMass = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/basicMass"][:]
        indexHost = np.argmax(basicMass)

        densityProfileHost = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/densityProfile"][indexHost]

        self.fileRead.close()
        return densityProfileHost

    def density(self, i):
        # Return the density profile of the subhalo.
        self.fileRead = h5py.File(self.fileName, 'r')
        basicMass = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/basicMass"][:]
        indexSat = np.argmin(basicMass)

        densityProfile = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/densityProfile"][indexSat]

        self.fileRead.close()
        return densityProfile

    def densityProj(self, i):
        # Return the projected density of the subhalo.
        self.fileRead = h5py.File(self.fileName, 'r')
        basicMass = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/basicMass"][:]
        indexSat = np.argmin(basicMass)

        densityProfileProj = self.fileRead["Outputs/Output" + str(i + 1) + "/nodeData/projectedDensity"][indexSat]

        self.fileRead.close()
        return densityProfileProj




import numpy as np
import numpy.testing as npt
import pytest
from pyHalo.Rendering.SpatialDistributions.correlated import Correlated2D

class GeometryDummy(object):

    def rendering_scale(self, *args, **kwargs):

        return 1.

    def kpc_per_arcsec(self, *args,  **kwargs):

        return 1.

class TestCorrelated(object):

    def setup_method(self):

        geometry = GeometryDummy()
        self.correlated = Correlated2D(geometry, 1.)

    def test_correlated(self):

        sigma = 0.8
        r_max = 3. * sigma
        n_points = 500000
        nbins = 200

        mean = [0., 0.]
        cov = [[sigma**2, 0], [0, sigma**2]]
        out = np.random.multivariate_normal(mean, cov, n_points)
        x_gaussian, y_gaussian = out[:,0], out[:, 1]
        r_theory = np.hypot(x_gaussian, y_gaussian)
        density, _,  _ = np.histogram2d(x_gaussian, y_gaussian, bins=nbins,
                                        range=((-r_max, r_max), (-r_max, r_max)))

        x, y = self.correlated.draw(n_points, r_max, density, 1.)
        r = np.hypot(x, y)
        h, _, _ = np.histogram2d(x, y, bins=100, range=((-r_max, r_max), (-r_max, r_max)))
        h_theory, _, _ = np.histogram2d(x_gaussian, y_gaussian, bins=100, range=((-r_max, r_max), (-r_max, r_max)))

        radii = np.linspace(0, 1. * r_max, 50)
        rendered = []
        theory = []
        for i in range(0, len(radii)-1):
            rad = radii[i + 1]**2 - radii[i]**2

            condition = np.logical_and(r >= radii[i], r < radii[i+1])
            n_rendered = np.sum(np.where(condition))
            rendered.append(n_rendered/rad)

            condition = np.logical_and(r_theory >= radii[i], r_theory < radii[i + 1])
            n_theory = np.sum(np.where(condition))
            theory.append(n_theory / rad)

        residuals = np.absolute(1-np.array(theory)/np.array(rendered))
        mean_error = np.mean(residuals)
        npt.assert_equal(len(x), len(y))
        npt.assert_equal(len(x), n_points)
        npt.assert_equal(True, mean_error < 0.05)


if __name__ == '__main__':

    pytest.main()



import numpy as np
from pyHalo.Rendering.MassFunctions.util import integrate_power_law_analytic, integrate_power_law_quad
from pyHalo.utilities import inverse_transform_sampling
from scipy.interpolate import interp1d

__all__ = ['CDMPowerLaw', 'WDMPowerLaw', 'MixedWDMPowerLaw']


class _PowerLawBase(object):
    name = 'BASE_POWER_LAW'
    def __init__(self, log_mlow, log_mhigh, power_law_index, draw_poisson, normalization):
        """
        :param log_mlow: log10(minimum halo mass)
        :param log_mhigh: log10(maximum halo mass)
        :param power_law_index: logarithmic slope of the halo mass function between mlow and mhigh
        :param draw_poisson: bool; whether to draw number of halos from a poisson distribution
        :param normalization: the amplitude of the mass function
        """
        self._logmlow = log_mlow
        self._logmhigh = log_mhigh
        self._index = power_law_index
        self._draw_poisson = draw_poisson
        self._normalization = normalization
        self.n_mean = integrate_power_law_analytic(normalization, 10 ** log_mlow, 10**log_mhigh,
                                                   0.0, power_law_index)
        self.first_moment = integrate_power_law_analytic(normalization, 10 ** log_mlow, 10**log_mhigh,
                                                   1.0, power_law_index)

    def draw(self):

        """
        Sample from the mass function
        """

        mH = 10 ** self._logmhigh
        mL = 10 ** self._logmlow
        if self._draw_poisson:
            ndraw = np.random.poisson(self.n_mean)
        else:
            ndraw = int(round(np.round(self.n_mean)))
        x = np.random.uniform(0, 1, ndraw)
        if self._index == -1:
            norm = np.log(mH / mL)
            X = mL * np.exp(norm * x)
        else:
            X = (x * (mH ** (1 + self._index) - mL ** (1 + self._index)) + mL ** (
                1 + self._index)) ** (
                    (1 + self._index) ** -1)
        return np.array(X)

class _PowerLawTurnoverBase(object):
    name = 'BASE_POWER_LAW_TURNOVER'
    def __init__(self, log_mlow, log_mhigh, power_law_index, draw_poisson,
                 normalization, kwargs_mass_function):
        """

        :param log_mlow: log10(minimum halo mass)
        :param log_mhigh: log10(maximum halo mass)
        :param log_mc: log10(halo mass) where turnover occurs
        :param power_law_index: logarithmic slope of the halo mass function between mlow and mhigh
        :param draw_poisson: bool; whether to draw number of halos from a poisson distribution
        :param normalization: the amplitude of the mass function
        :param a_mfunc_break: coefficient of turnover
        :param b_mfunc_break: exponent of turnover
        :param c_mfunc_break: outere exponent of turnover
        """
        self._logmlow = log_mlow
        self._logmhigh = log_mhigh
        self._index = power_law_index
        self._draw_poisson = draw_poisson
        self._normalization = normalization
        self._kwargs_mass_function = kwargs_mass_function
        self._mass_function_unbroken = _PowerLawBase(log_mlow, log_mhigh, power_law_index, draw_poisson, normalization)
        self.n_mean = integrate_power_law_quad(normalization, 10 ** log_mlow, 10 ** log_mhigh,
                                                         0.0, power_law_index, self._turnover,
                                                     self._kwargs_mass_function)

        self.first_moment = integrate_power_law_quad(normalization, 10 ** log_mlow, 10 ** log_mhigh,
                                                         1.0, power_law_index, self._turnover,
                                                     self._kwargs_mass_function)

    def draw(self):
        """
        Samples from the mass function
        :return:
        """
        m = self._mass_function_unbroken.draw()
        if len(m) == 0:
            return m
        factor = self._turnover(m, **self._kwargs_mass_function)
        u = np.random.rand(int(len(m)))
        inds = np.where(u < factor)
        return m[inds]

    @staticmethod
    def _turnover(*args, **kwargs):
        raise Exception('must specify turnover method')


class CDMPowerLaw(_PowerLawBase):
    name = 'CDM_POWER_LAW'
    def __init__(self, log_mlow, log_mhigh, power_law_index, draw_poisson, normalization, *args, **kwargs):
        """

       :param log_mlow: log10(minimum halo mass)
       :param log_mhigh: log10(maximum halo mass)
       :param power_law_index: logarithmic slope of the halo mass function between mlow and mhigh
       :param draw_poisson: bool; whether to draw number of halos from a poisson distribution
       :param normalization: the amplitude of the mass function
       """
        super(CDMPowerLaw, self).__init__(log_mlow, log_mhigh, power_law_index, draw_poisson, normalization)


class WDMPowerLaw(_PowerLawTurnoverBase):
    name = 'WDM_POWER_LAW'
    def __init__(self, log_mlow, log_mhigh, power_law_index, draw_poisson, normalization,
                 log_mc, a_wdm, b_wdm, c_wdm, *args, **kwargs):
        """

       :param log_mlow: log10(minimum halo mass)
       :param log_mhigh: log10(maximum halo mass)
       :param power_law_index: logarithmic slope of the halo mass function between mlow and mhigh
       :param draw_poisson: bool; whether to draw number of halos from a poisson distribution
       :param normalization: the amplitude of the mass function
       :param log_mc: log10(halo mass) where turnover occurs
       :param a_wdm: coefficient of turnover
       :param b_wdm: exponent of turnover
       :param c_wdm: outer exponent of turnover
       """

        kwargs_mass_function = {'log_mc': log_mc, 'a_mfunc_break': a_wdm, 'b_mfunc_break': b_wdm,
                                'c_mfunc_break': c_wdm}
        super(WDMPowerLaw, self).__init__(log_mlow, log_mhigh, power_law_index, draw_poisson, normalization,
                                          kwargs_mass_function)

    def draw(self):
        """
        Samples from the mass function
        :return:
        """
        m = self._mass_function_unbroken.draw()
        if len(m) == 0:
            return m
        factor = self._turnover(m, **self._kwargs_mass_function)
        u = np.random.rand(int(len(m)))
        inds = np.where(u < factor)
        return m[inds]

    @staticmethod
    def _turnover(m, log_mc, a_mfunc_break, b_mfunc_break, c_mfunc_break):
        """

        :param m:
        :param log_mc:
        :param a_mfunc_break:
        :param b_mfunc_break:
        :param c_mfunc_break:
        :return:
        """
        m_c = 10 ** log_mc
        r = a_mfunc_break * (m_c / m) ** b_mfunc_break
        factor = 1 + r
        return factor ** c_mfunc_break

class MixedWDMPowerLaw(_PowerLawTurnoverBase):
    name = 'MIXED_DM_POWER_LAW'
    def __init__(self, log_mlow, log_mhigh, power_law_index, draw_poisson, normalization,
                 log_mc, a_wdm, b_wdm, c_wdm, mixed_DM_frac, *args, **kwargs):
        """

       :param log_mlow: log10(minimum halo mass)
       :param log_mhigh: log10(maximum halo mass)
       :param power_law_index: logarithmic slope of the halo mass function between mlow and mhigh
       :param draw_poisson: bool; whether to draw number of halos from a poisson distribution
       :param normalization: the amplitude of the mass function
       :param log_mc: log10(halo mass) where turnover occurs
       :param a_wdm: coefficient of turnover
       :param b_wdm: exponent of turnover
       :param c_wdm: outer exponent of turnover
       :param mixed_DM_frac: fraction of dark matter in CDM component
       """
        kwargs_mass_function = {'log_mc': log_mc, 'a_mfunc_break': a_wdm, 'b_mfunc_break': b_wdm,
                                'c_mfunc_break': c_wdm, 'mixed_DM_frac': mixed_DM_frac}
        super(MixedWDMPowerLaw, self).__init__(log_mlow, log_mhigh, power_law_index, draw_poisson,
                                               normalization, kwargs_mass_function)

    @staticmethod
    def _turnover(m, log_mc, a_mfunc_break, b_mfunc_break, c_mfunc_break, mixed_DM_frac):
        """

        :param m:
        :param log_mc:
        :param a_mfunc_break:
        :param b_mfunc_break:
        :param c_mfunc_break:
        :param mixed_DM_frac:
        :return:
        """
        m_c = 10 ** log_mc
        r = a_mfunc_break * (m_c / m) ** b_mfunc_break
        factor = 1 + r
        wdm_comp = factor ** c_mfunc_break
        suppression_factor = (mixed_DM_frac + (1 - mixed_DM_frac) * np.sqrt(wdm_comp)) ** 2
        return suppression_factor

# class Tabulated(object):
#     """
#     This class generates samples from an mass function given an array of halo masses and the amplitude of the differential
#     mass function at these masses
#     """
#     def __init__(self, log_mlow, log_mhigh, draw_poisson, m, dndm):
#         """
#
#         :param log_mlow: minimum halo mass to render
#         :param log_mhigh: maximum halo mass to render
#         :param draw_poisson: bool; draw from a poisson distribution
#         :param m: numpy array containing halo masses
#         :param dndm: differential mass function amplitudes corresponding to m
#         """
#
#         self._logmlow = log_mlow
#         self._logmhigh = log_mhigh
#         self._draw_poisson = draw_poisson
#         cond1 = m >= 10 ** log_mlow
#         cond2 = m <= 10 ** log_mhigh
#         inds_keep = np.where(np.logical_and(cond1, cond2))[0]
#         self._m = m[inds_keep]
#         self._dndm = dndm[inds_keep]
#         self.n_mean = np.trapz(self._dndm, self._m)
#         self.first_moment = np.trapz(self._m * self._dndm, self._m)
#         self._interp_dndm = interp1d(self._m, self._dndm)
#
#     def draw(self):
#         """
#         Samples from the mass function
#         :return: an array of masses
#         """
#         if self._draw_poisson:
#             n_draw = int(np.random.poisson(self.n_mean))
#         else:
#             n_draw = int(self.n_mean)
#
#         samples = inverse_transform_sampling(self._m, self._interp_dndm, (), n_draw)
#         return samples




from lenstronomy.LensModel.lens_model import LensModel
import numpy.testing as npt
from scipy.integrate import simpson as simps
import numpy as np
from pyHalo.utilities import interpolate_ray_paths, de_broglie_wavelength, delta_sigma, ITSampling, \
    inverse_transform_sampling, delta_kappa, nfw_velocity_dispersion
from pyHalo.Cosmology.cosmology import Cosmology
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.utilities import mask_annular
import pytest

class TestUtilities(object):

    def test_interp_ray_paths(self):

        cosmo = Cosmology()
        x = [1.4, -1.]
        y = [-0.4, 0.2]
        lens_model = LensModel(['SIS'], z_source=2., lens_redshift_list=[0.5], multi_plane=True)
        kwargs_lens = [{'theta_E': 1., 'center_x': 0., 'center_y': 0.}]

        interpx, interpy = interpolate_ray_paths(x, y, lens_model, kwargs_lens, 2.,
                                                             terminate_at_source=False, source_x=None, source_y=None,
                                                             evaluate_at_mean=False)
        zarray = np.linspace(0., 2., 100)
        dc = [cosmo.D_C_transverse(zi) for zi in zarray]
        interpx_mean, interpy_mean = interpolate_ray_paths(x, y, lens_model, kwargs_lens, 2.,
                                                                       terminate_at_source=False, source_x=None,
                                                                       source_y=None,
                                                                       evaluate_at_mean=True, cosmo=cosmo)
        for dci in dc:
            x, y = 0.5 * (interpx[0](dci) + interpx[1](dci)), 0.5 * (interpy[0](dci) + interpy[1](dci))
            npt.assert_almost_equal(x, interpx_mean[0](dci))
            npt.assert_almost_equal(y, interpy_mean[0](dci))

        x = [1.4, -1.]
        y = [-0.4, 0.2]
        source_x, source_y = 0.2, 0.5
        interpx, interpy = interpolate_ray_paths(x, y, lens_model, kwargs_lens, 2.,
                                                             terminate_at_source=True, source_x=source_x,
                                                             source_y=source_y,
                                                             evaluate_at_mean=True)
        npt.assert_almost_equal(interpx[0](dc[-1]), source_x)
        npt.assert_almost_equal(interpy[0](dc[-1]), source_y)

    def test_uldm_functions(self):

        log10_m_uldm=-22 # log(eV)
        v=200 # km/s
        z_lens,z_source=0.5,1.5
        m=1e13 #M_solar
        rein=6. #kpc
        lens_cosmo = LensCosmo(z_lens, z_source)
        lambda_dB = de_broglie_wavelength(log10_m_uldm,v)
        npt.assert_almost_equal(lambda_dB,0.6)
        delta_s = delta_sigma(m,z_lens,rein,lambda_dB)
        npt.assert_almost_equal(delta_s/80837585.696, 1, 2)
        dk = delta_kappa(z_lens, z_source, 10 ** 13.0, rein, lambda_dB)
        sigma_crit = lens_cosmo.get_sigma_crit_lensing(z_lens, z_source) * (1e-3) ** 2
        npt.assert_almost_equal(delta_s, dk * sigma_crit)

    def test_nfw_velocity_dispersion(self):

        lens_comso = LensCosmo(0.5, 1.5)
        m = 10**8
        z = 0.5
        c = 15.0
        rhos, rs, _ = lens_comso.NFW_params_physical(m, c, z)
        vdis1 = nfw_velocity_dispersion(rhos, rs, c)
        npt.assert_almost_equal(vdis1/6.34, 1.0, 2)

    def test_inverse_transform_sampling(self):

        mu = 2.0
        sigma = 1.5
        func = lambda x: np.exp(-0.5 * (x - mu) ** 2 / sigma**2)
        x = np.linspace(mu - 5*sigma, mu + 5*sigma, 1000)
        x_samples = inverse_transform_sampling(x, func, (), 100000)
        npt.assert_almost_equal(np.mean(x_samples)/mu, 1.0, 2)
        npt.assert_almost_equal(np.std(x_samples)/sigma, 1.0, 2)

    def test_ITSampling(self):

        mu = 2.14
        sigma = 0.35
        x = np.random.normal(mu, sigma, 250000)
        sampler = ITSampling.from_samples(x)
        x_samples = sampler(100000)
        npt.assert_almost_equal(np.mean(x_samples)/mu, 1.0, 2)
        npt.assert_almost_equal(np.std(x_samples)/sigma, 1.0, 2)

    def test_mask_annular(self):
        npix = 1000
        window_size = 4
        delta_pix = window_size/npix
        _R = np.linspace(-window_size/2, window_size/2, npix)
        XX, YY = np.meshgrid(_R, _R)
        r_min, r_max = 0.5, 1.5
    
        no_mask = mask_annular(0, 0, XX, YY, r_min = 0, r_max = None)
        mask_1 = mask_annular(0, 0, XX, YY, r_min, r_max = None)
        mask_2 = mask_annular(0, 0, XX, YY, r_min, r_max)
    
        area_no_mask = simps(simps(no_mask, _R, axis=0), _R, axis=-1)
        area_mask_1 = simps(simps(mask_1, _R, axis=0), _R, axis=-1)
        area_mask_2 = simps(simps(mask_2, _R, axis=0), _R, axis=-1)
    
        area_no_mask_real = window_size**2
        area_mask_1_real = window_size**2 - np.pi*r_min**2
        area_mask_2_real = np.pi*(r_max**2 - r_min**2)
    
        npt.assert_array_almost_equal(area_no_mask_real,area_no_mask, decimal=3)
        npt.assert_array_almost_equal(area_mask_1_real,area_mask_1, decimal=3)
        npt.assert_array_almost_equal(area_mask_2_real,area_mask_2, decimal=3)

if __name__ == '__main__':

    pytest.main()



import pytest
import numpy as np
import numpy.testing as npt
from pyHalo.Rendering.two_halo import TwoHaloContribution, two_halo_enhancement_factor
from pyHalo.Rendering.MassFunctions.density_peaks import ShethTormen
from pyHalo.Cosmology.cosmology import Cosmology
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.utilities import generate_lens_plane_redshifts
from pyHalo.Rendering.SpatialDistributions.uniform import LensConeUniform


class TestTwoHalo(object):

    def setup_method(self):

        self.zlens = 0.5
        self.zsource = 1.5
        self.cosmo = Cosmology()
        self.lens_cosmo = LensCosmo(self.zlens, self.zsource, self.cosmo)

        mass_function_model = ShethTormen
        kwargs_mass_function = {'log_mlow': 6.0, 'log_mhigh': 10.0, 'm_pivot': 1e8,
                                'LOS_normalization': 1.0, 'delta_power_law_index': 0.0, 'log_m_host': 13.0,
                                'draw_poisson': False}

        geometry = Geometry(self.cosmo, self.zlens, self.zsource, 6.0, 'DOUBLE_CONE')
        lens_cosmo = LensCosmo(self.zlens, self.zsource, self.cosmo)
        lens_plane_redshifts, delta_z_list = generate_lens_plane_redshifts(self.zlens, self.zsource)
        spatial_distribution_model = LensConeUniform(geometry.cone_opening_angle, geometry)

        self.model = TwoHaloContribution(mass_function_model, kwargs_mass_function,
                                                       spatial_distribution_model, geometry, lens_cosmo,
                                                       lens_plane_redshifts, delta_z_list)

        use_Lazar_correction = False
        self.model_no_Lazar = TwoHaloContribution(mass_function_model, kwargs_mass_function,
                                         spatial_distribution_model, geometry, lens_cosmo,
                                         lens_plane_redshifts, delta_z_list, use_Lazar_correction)

    def test_name(self):

        npt.assert_string_equal('TWO_HALO_TERM', self.model.name)

    def test_boost(self):

        z_step = 0.02
        mhost = 10**13
        r200_host = 0.4
        boost = two_halo_enhancement_factor(self.zlens, z_step, self.lens_cosmo, mhost, r200_host,
                                            use_Lazar_correction=True)
        npt.assert_equal(boost>0, True)
        boost_no_lazar = two_halo_enhancement_factor(self.zlens, z_step, self.lens_cosmo, mhost, r200_host,
                                                       use_Lazar_correction=False)
        npt.assert_equal(boost_no_lazar / boost < 1, True)

    def test_render(self):

        masses, x, y, r3d, redshifts, subhalo_flag = self.model.render()
        npt.assert_equal(len(masses), len(x))
        npt.assert_equal(len(masses), len(y))
        npt.assert_equal(len(masses), len(r3d))
        npt.assert_equal(len(masses), len(redshifts))
        npt.assert_equal(len(masses), len(subhalo_flag))
        for flag in subhalo_flag:
            npt.assert_equal(False, flag)
        npt.assert_equal(redshifts, self.zlens)

        masses_no_Lazar, x, y, r3d, redshifts, subhalo_flag = self.model_no_Lazar.render()
        npt.assert_equal(len(masses_no_Lazar) < len(masses), True)

if __name__ == '__main__':
   pytest.main()



import numpy as np
from scipy.interpolate import RegularGridInterpolator
from scipy.stats import johnsonsu

class InterpGalacticusKeeley24(object):
    """
    This class interpolates output from the semi-analytic model galacticus to predict the bound mass of a subhalo
    as a function of its infall mass, concentration, host halo concentration, and the time since infall.
    """
    def __init__(self):
        from pyHalo.Halos.galacticus_truncation.tabulated_mass_loss_Keeley24 import _log10_mbound_over_minfall, \
            _log10_mbound_over_minfall_scatter_dex
        _t_ellapsed_min, _t_ellapsed_max = 0.0, 7.658449372663094
        _chostmin, _chostmax = 2.0, 8.0
        _log10cmin, _log10cmax = 0.3010299956639812, 2.3010299956639813
        _n = 20
        _t_coords = np.linspace(_t_ellapsed_min, _t_ellapsed_max, _n)
        _chost_coords = np.linspace(_chostmin, _chostmax, _n)
        _log10c_coords = np.linspace(_log10cmin, _log10cmax, _n)
        _values = np.array(_log10_mbound_over_minfall).reshape(_n, _n, _n)
        _values_scatter = np.array(_log10_mbound_over_minfall_scatter_dex).reshape(_n, _n, _n)
        _points = (_t_coords, _chost_coords, _log10c_coords)
        self._mfrac_interp = RegularGridInterpolator(_points, _values, bounds_error=False, fill_value=None)
        self._mfrac_scatter_dex_interp = RegularGridInterpolator(_points, _values_scatter, bounds_error=False,
                                                                 fill_value=None)

    def __call__(self, log10_concentration_infall, time_since_infall, host_concentration):
        """
        Evaluates the prediction from galacticus for subhalo bound mass
        :param log10_concentration_infall: log10(c) where c is the halo concentration at infall
        :param time_since_infall: the time ellapsed since infall and the deflector redshift
        :param host_concentration: the concentration of the host halo
        :return: the log10(bound mass divided by the infall mass), plus scatter
        """
        mean = self.evaluate_mean_mass_loss(log10_concentration_infall, time_since_infall, host_concentration)
        scatter_dex = self.evaluate_scatter_dex(log10_concentration_infall, time_since_infall, host_concentration)
        scatter_dex = max(scatter_dex, 0.001)
        output = np.random.normal(mean, scatter_dex)
        return output

    def evaluate_mean_mass_loss(self, log10_concentration_infall, time_since_infall, host_concentration):
        """

        :param log10_concentration_infall: log10(c) where c is the halo concentration at infall
        :param time_since_infall: the time ellapsed since infall and the deflector redshift
        :param host_concentration: the concentration of the host halo
        :return: log10 bound mass divided by the infall mass
        """
        point = (time_since_infall, host_concentration, log10_concentration_infall)
        y = self._mfrac_interp(point)
        if isinstance(log10_concentration_infall, float) and \
            isinstance(time_since_infall, float) and \
            isinstance(host_concentration, float):
            return float(y)
        else:
            return np.squeeze(y)

    def evaluate_scatter_dex(self, log10_concentration_infall, time_since_infall, host_concentration):
        """

        :param log10_concentration_infall: log10(c) where c is the halo concentration at infall
        :param time_since_infall: the time ellapsed since infall and the deflector redshift
        :param host_concentration: the concentration of the host halo
        :return: the scatter in dex of the bound mass divided by infall mass
        """
        point = (time_since_infall, host_concentration, log10_concentration_infall)
        y = self._mfrac_scatter_dex_interp(point)
        if isinstance(log10_concentration_infall, float) and \
            isinstance(time_since_infall, float) and \
            isinstance(host_concentration, float):
            return float(y)
        else:
            return np.squeeze(y)

class InterpGalacticus(object):
    """
    This class interpolates output from the semi-analytic model galacticus to predict the bound mass of a subhalo
    as a function of its infall mass, concentration, host halo concentration, and the time since infall.
    """
    def __init__(self):
        from pyHalo.Halos.galacticus_truncation.johnsonSUparams import a_fit, \
            b_fit
        log10c_values = np.linspace(np.log10(2.0), np.log10(384), 25)
        t_inf_values = np.linspace(0.0, 8.1, 25)
        chost_values = np.linspace(4.0, 9.0, 25)
        a_values = np.array(a_fit).reshape(25, 25, 25)
        b_values = np.array(b_fit).reshape(25, 25, 25)
        _points = (t_inf_values, log10c_values, chost_values)
        self._a_interp = RegularGridInterpolator(_points, a_values, bounds_error=False,
                                               fill_value=None)
        self._b_interp = RegularGridInterpolator(_points, b_values, bounds_error=False,
                                           fill_value=None)

    def __call__(self, log10_concentration_infall, time_since_infall, chost):
        """
        Evaluates the prediction from galacticus for subhalo bound mass
        :param log10_concentration_infall: log10(c) where c is the halo concentration at infall
        :param time_since_infall: the time ellapsed since infall and the deflector redshift
        :param chost: host halo concentration at z=0.5
        :return: the log10(bound mass divided by the infall mass), plus scatter
        """

        log10_concentration_infall = max(np.log10(2), log10_concentration_infall)
        log10_concentration_infall = min(np.log10(384), log10_concentration_infall)
        time_since_infall = max(0.0, time_since_infall)
        time_since_infall = min(time_since_infall, 8.1)
        chost = max(4.0, chost)
        chost = min(9.0, chost)
        p = (time_since_infall, log10_concentration_infall, chost)
        a, b = self._a_interp(p), self._b_interp(p)
        output = float(johnsonsu.rvs(a, b))
        return output
